from ast import Mult
import unittest

from rockit import Ocp, DirectMethod, MultipleShooting, FreeTime, DirectCollocation, SingleShooting, SplineMethod, UniformGrid, GeometricGrid, FreeGrid, LseGroup, rockit_pickle_context, rockit_unpickle_context
from problems import integrator_control_problem, vdp, vdp_dae, bang_bang_problem
from casadi import DM, jacobian, sum1, sum2, MX, rootfinder, evalf, sumsqr, symvar, vertcat
import casadi as ca
from numpy import sin, pi, linspace
from numpy.testing import assert_array_almost_equal
from rockit.splines.spline import Spline
try:
  from contextlib import redirect_stdout
except:
  redirect_stdout = None
from io import StringIO
import numpy as np

class MiscTests(unittest.TestCase):

    def test_grid_convergence(self):
        f_exact = 4.414113817112848

        DM.set_precision(16)
        fss = []
        for method,order in [
          (lambda M: MultipleShooting(N=6,intg='collocation',intg_options={"number_of_finite_elements": M, "interpolation_order": 4}),6),
          (lambda M: MultipleShooting(N=6,intg='rk',M=M), 4),
          (lambda M: MultipleShooting(N=6,intg='cvodes',intg_options={"reltol":10**(-M),"abstol":10**(-M)}),2),
          (lambda M: DirectCollocation(N=6,M=M),7)]:
          fs = []
          for M in [1,2,4,8]:

            ocp, x1, x2, u = vdp(method(M))
            ocp.solver("ipopt",{"ipopt.tol":1e-14})
            sol = ocp.solve()


            f = sol.value(ocp.objective)
            fs.append(f_exact-f)

          print(fs)
          # An increase of factor 2 in M results in a factor 2**order in objective accuracy
          order_meas = -np.log(np.abs(np.array(fs[1:])/np.array(fs[:-1])))/np.log(2)
          print(order_meas)
          self.assertTrue(np.max(order_meas)>order-0.01)

    def test_spy(self):
      ocp, _, _ = integrator_control_problem()
      ocp.solve()
      ocp.spy()
      import matplotlib.pylab as plt
      self.assertEqual(plt.gca().title._text, "Lagrange Hessian: 101x101,0nz")

    def test_ocp_objective(self):
      ocp, x, u = integrator_control_problem()
      sol = ocp.solve()
      self.assertAlmostEqual(sol.value(ocp.objective),sol.value(ocp.at_tf(x)))

    def test_splinemethod_order0(self):
      for method in [MultipleShooting(N=10),SplineMethod(N=10)]:
         for order in [0,1]:
            # Only control
            ocp = Ocp(T=1)

            x = ocp.control(order=order)
            ocp.add_objective(ocp.sum(sumsqr(x-ocp.t),include_last=True))
            ocp.solver('ipopt')

            ocp.method(method)

            # Solve
            sol = ocp.solve()
            [t,xs] = sol.sample(x, grid='control')
            if order==1:
              assert_array_almost_equal(t, xs)
            if order==0:
              assert_array_almost_equal(t[:-2], xs[:-2])
              assert_array_almost_equal(xs[-2:],0.95)

            if "SplineMethod" in str(method):
              print(sol.sample(x, grid='gist'))

            ocp = Ocp(T=1)

            # Only variables
            ocp = Ocp(T=1)

            x = ocp.variable(grid='bspline',order=order)
            ocp.add_objective(ocp.sum(sumsqr(x-ocp.t),include_last=True))
            ocp.solver('ipopt')

            ocp.method(method)

            # Solve
            sol = ocp.solve()
            [t,xs] = sol.sample(x, grid='control')
            if order==1:
              assert_array_almost_equal(t, xs)
            if order==0:
              assert_array_almost_equal(t[:-2], xs[:-2])
              assert_array_almost_equal(xs[-2:],0.95)


    def test_grid_gist(self):
      for order in range(3):
        ocp = Ocp(T=1)
        x = ocp.control(2,order=order)
        ocp.add_objective(ocp.sum(sumsqr(x-ocp.t),include_last=True))
        ocp.solver('ipopt')
        ocp.method(SplineMethod(N=10))
        sol = ocp.solve()

        [_,xs] = sol.sample(x, grid='gist')
        self.assertEqual(xs.shape[0], 10+order)
        [_,xs2] = sol.sample(7*x+1.7, grid='gist')
        assert_array_almost_equal(7*xs+1.7,xs2)
        [_,xs2] = sol.sample(vertcat(3*x[0]+1.3,x[1]+1.7), grid='gist')
        assert_array_almost_equal(xs[:,0]*3+1.3,xs2[:,0])
        assert_array_almost_equal(xs[:,1]+1.7,xs2[:,1])
        [_,xs2] = sol.sample(vertcat(x[1]+1.3,3*x[0]+1.7), grid='gist')
        assert_array_almost_equal(xs[:,1]+1.3,xs2[:,0])
        assert_array_almost_equal(xs[:,0]*3+1.7,xs2[:,1])
        [_,xs] = ocp.sample(x, grid='gist')
        self.assertEqual(xs.shape[1], 10+order)

        ocp = Ocp(T=1)
        x = ocp.variable(2,grid='bspline',order=order)
        ocp.add_objective(ocp.sum(sumsqr(x-ocp.t),include_last=True))
        ocp.solver('ipopt')
        ocp.method(SplineMethod(N=10))
        sol = ocp.solve()

        [_,xs] = sol.sample(x, grid='gist')
        self.assertEqual(xs.shape[0], 10+order)
        [_,xs2] = sol.sample(7*x+1, grid='gist')
        assert_array_almost_equal(7*xs+1,xs2)
        [_,xs] = ocp.sample(x, grid='gist')
        self.assertEqual(xs.shape[1], 10+order)

        ocp = Ocp(T=1)
        x = ocp.parameter(2,grid='bspline',order=order)
        DM.rng(1)
        ocp.set_value(x, DM.rand(2,10+order))
        ocp.add_objective(ocp.sum(sumsqr(x),include_last=True))
        ocp.solver('ipopt')
        ocp.method(SplineMethod(N=10))
        sol = ocp.solve()

        [_,xs] = sol.sample(x, grid='gist')
        self.assertEqual(xs.shape[0], 10+order)
        [_,xs2] = sol.sample(7*x+1, grid='gist')
        assert_array_almost_equal(7*xs+1,xs2)
        [_,xs] = ocp.sample(x, grid='gist')
        self.assertEqual(xs.shape[1], 10+order)

    def test_der_signals(self):
      for grid in [None,'inf']:
        # Only variables
        ocp = Ocp(T=1.3)

        x = ocp.variable(grid='bspline',order=2)
        ocp.subject_to(ocp.at_t0(x)==0)
        ocp.add_objective(-ocp.at_tf(x))

        ocp.subject_to(ocp.der(x)<=1.1,grid=grid)
        ocp.solver('ipopt')
        ocp.method(SplineMethod(N=10))
        sol = ocp.solve()

        assert_array_almost_equal(sol.sample(ocp.der(x),grid='gist')[1], 1.1)

        self.assertAlmostEqual(sol.value(ocp.at_tf(x)),1.3*1.1)

    def test_p_bspline(self):
      ocp = Ocp(T=1)
      x = ocp.control(order=2)

      p = ocp.parameter(grid='bspline',order=2)
      ocp.add_objective(ocp.sum(sumsqr(x-p),include_last=True))
      DM.rng(1)
      coeff = DM.rand(1,12)
      ocp.set_value(p, coeff)
      ocp.solver('ipopt')
      ocp.method(SplineMethod(N=10))

      # Solve
      sol = ocp.solve()
      [_,xs] = sol.sample(x, grid='gist')
      #print(xs.shape)
      #print(np.array(coeff).squeeze().shape)
      #assert_array_almost_equal(np.array(coeff).squeeze(), xs)

      [_,ds] = sol.sample(x-p, grid='control')
      assert_array_almost_equal(ds, 0)

    def test_basic(self):
        for T in [1, 3.2]:
            for M in [1, 2]:
                for u_max in [1, 2]:
                    for t0 in [0, 1]:
                        for x0 in [0, 1]:
                            for method in [MultipleShooting(N=4,M=M,intg='rk'),
                                           MultipleShooting(N=4,M=M,intg='cvodes'),
                                           MultipleShooting(N=4,M=M,intg='idas'),
                                           DirectCollocation(N=4,M=M),
                                           SingleShooting(N=4,M=M,intg='rk'),
                                           SplineMethod(N=4)]:
                                ocp, x, u = integrator_control_problem(
                                    T, u_max, x0, method, t0
                                )
                                sol = ocp.solve()

                                ts, xs = sol.sample(x, grid='control')

                                self.assertAlmostEqual(xs[0], x0, places=6)
                                self.assertAlmostEqual(
                                    xs[-1], x0 - u_max * T, places=6)
                                self.assertAlmostEqual(ts[0], t0)
                                self.assertAlmostEqual(ts[-1], t0 + T)

    def test_der(self):
        T = 1
        M = 1
        b = 1
        t0 = 0
        x0 = 0
        ocp = Ocp(t0=t0,T=T)

        x = ocp.state()
        u = ocp.control()

        ocp.set_der(x,u)

        y = 2*x

        ocp.subject_to(ocp.der(y)<=2*b)
        ocp.subject_to(-2*b<=ocp.der(y))
       
        ocp.add_objective(ocp.at_tf(x))
        ocp.subject_to(ocp.at_t0(x)==x0)

        ocp.solver('ipopt')

        ocp.method(MultipleShooting(N=4,M=M,intg='rk'))
       
        sol = ocp.solve()

        ts, xs = sol.sample(x,grid='control')

        self.assertAlmostEqual(xs[0],x0,places=6)
        self.assertAlmostEqual(xs[-1],x0-b*T,places=6)
        self.assertAlmostEqual(ts[0],t0)
        self.assertAlmostEqual(ts[-1],t0+T)

    def test_parametric_time(self):
        ocp = Ocp()

        T = ocp.parameter()
        ocp.set_T(T)
        ocp.set_value(T, 3)
        x = ocp.state()
        u = ocp.control()

        ocp.set_der(x, u)

        ocp.subject_to(u <= 1)
        ocp.subject_to(-1 <= u)

        ocp.add_objective(ocp.at_tf(x))
        ocp.subject_to(ocp.at_t0(x) == 0)

        ocp.solver('ipopt')

        ocp.method(MultipleShooting())

        sol = ocp.solve()

        ts, xs = sol.sample(x, grid='control')
        self.assertAlmostEqual(xs[0], 0)
        self.assertAlmostEqual(xs[-1], -3,places=6)
        self.assertAlmostEqual(ts[-1], 3,places=6)

        ocp.set_value(T, 1)
        sol = ocp.solve()

        ts, xs = sol.sample(x, grid='control')
        self.assertAlmostEqual(xs[-1], -1,places=6)
        self.assertAlmostEqual(ts[-1], 1,places=6)
  
    def test_basic_time_free(self):
        xf = 2
        for t0 in [0, 1]:
            for x0 in [0, 1]:
                for b in [1.0, 2.0]:
                    for method in [MultipleShooting(N=4, intg='rk'), MultipleShooting(N=4, intg='cvodes'), MultipleShooting(N=4, intg='idas'), DirectCollocation(N=4), SplineMethod(N=4)]:
                        for pos in ["pre","post"]:
                          if pos=="pre":
                            ocp = Ocp(t0=t0, T=FreeTime(1))
                          else:
                            ocp = Ocp(t0=t0)

                          x = ocp.state()
                          u = ocp.control()

                          ocp.set_der(x, u)

                          ocp.subject_to(u <= b)
                          ocp.subject_to(-b <= u)

                          ocp.add_objective(ocp.T)
                          ocp.subject_to(ocp.at_t0(x) == x0)
                          ocp.subject_to(ocp.at_tf(x) == xf)

                          ocp.solver('ipopt')

                          ocp.method(method)

                          if pos=="post":
                            ocp.set_T(ocp.variable())
                            ocp.set_initial(ocp.T, 1)
                            ocp.subject_to(ocp.T>=0)

                          sol = ocp.solve()

                          self.assertAlmostEqual(sol.value(ocp.t0), t0)
                          self.assertAlmostEqual(sol.value(ocp.T),(xf-x0)/b)
                          self.assertAlmostEqual(sol.value(ocp.tf), t0 + (xf - x0) / b)

                          # issue #91
                          self.assertAlmostEqual(sol.value(ocp.at_t0(x)), x0)
                          self.assertAlmostEqual(sol.value(ocp.at_tf(x)), xf)

                          ts, xs = sol.sample(x, grid='control')

                          self.assertAlmostEqual(xs[0], x0, places=6)
                          self.assertAlmostEqual(xs[-1], xf, places=6)
                          self.assertAlmostEqual(ts[0], t0)
                          self.assertAlmostEqual(ts[-1], t0 + (xf - x0) / b)

    def test_basic_t0_free(self):
        xf = 2
        t0 = 0
        for T in [2]:
            for x0 in [0, 1]:
                for b in [1, 2]:
                    for method in [MultipleShooting(N=4, intg='rk'), MultipleShooting(N=4, intg='cvodes'), MultipleShooting(N=4, intg='idas'), DirectCollocation(N=4), SplineMethod(N=4)]:
                        ocp = Ocp(t0=FreeTime(2),T=T)

                        x = ocp.state()
                        u = ocp.control()

                        ocp.set_der(x, u)
                        ocp.subject_to(u <= b)
                        ocp.subject_to(-b <= u)

                        ocp.add_objective(ocp.tf)
                        ocp.subject_to(ocp.at_t0(x) == x0)
                        ocp.subject_to(ocp.at_tf(x) == xf)
                        ocp.subject_to(ocp.t0 >= 0)

                        ocp.solver('ipopt')

                        ocp.method(method)

                        sol = ocp.solve()

                        ts, xs = sol.sample(x, grid='control')

                        self.assertAlmostEqual(xs[0], x0, places=6)
                        self.assertAlmostEqual(xs[-1], xf, places=6)
                        self.assertAlmostEqual(ts[0], t0)
                        self.assertAlmostEqual(ts[-1], t0 + T)

    def test_param(self):
      ocp = Ocp(T=1)

      x = ocp.state()
      u = ocp.control()

      p = ocp.parameter()

      ocp.set_der(x, u)

      ocp.subject_to(u <= 1)
      ocp.subject_to(-1 <= u)

      ocp.add_objective(ocp.at_tf(x))
      ocp.subject_to(ocp.at_t0(x) == p)

      ocp.solver('ipopt')

      ocp.method(MultipleShooting())

      ocp.set_value(p, 0)
      sol = ocp.solve()

      ts, xs = sol.sample(x, grid='control')
      self.assertAlmostEqual(xs[0], 0)

      ocp.set_value(p, 1)
      sol = ocp.solve()

      ts, xs = sol.sample(x, grid='control')
      self.assertAlmostEqual(xs[0], 1)

    def test_initial(self):
      for stage_method in [SingleShooting(),MultipleShooting(), DirectCollocation(), SplineMethod()]:
        ocp, x, u = integrator_control_problem(x0=None,stage_method=stage_method)

        ocp = Ocp(T=1)

        x = ocp.state()
        u = ocp.control()
        if "SplineMethod" in str(stage_method):
           ocp.set_der(x, u)
        else:
          p = ocp.parameter()
          ocp.set_value(p,1)
          ocp.set_der(x, p*u)

        ocp.subject_to(-1 <= (u <= 1))

        ocp.add_objective(ocp.at_tf(sin(x))+ocp.integral(u**2))

        ocp.solver('ipopt')

        ocp.method(stage_method)

        ocp.set_initial(x, -pi/2+0.1)
        sol = ocp.solve()
        ts, xs = sol.sample(x, grid='control')
        self.assertAlmostEqual(xs[0], -pi/2, places=6)
        ocp.set_initial(x, 2*pi-pi/2+0.1)
        sol = ocp.solve()
        ts, xs = sol.sample(x, grid='control')
        self.assertAlmostEqual(xs[0], 2*pi-pi/2, places=6)

      for stage_method in [SingleShooting(),MultipleShooting(), DirectCollocation(), SplineMethod()]:
        ocp, x, u = integrator_control_problem(x0=None,stage_method=stage_method)
        v = ocp.variable()
        ocp.subject_to(ocp.at_t0(x)==v)
        ocp.subject_to(0==sin(v))
        sol = ocp.solve()
        ts, xs = sol.sample(x, grid='control')
        self.assertAlmostEqual(xs[0], 0, places=6)

        ocp.set_initial(v, 2*pi)
        sol = ocp.solve()
        ts, xs = sol.sample(x, grid='control')
        self.assertAlmostEqual(xs[0], 2*pi, places=6)

    def test_show_infeasibilities(self):
      for method in [MultipleShooting(), DirectCollocation()]:
        ocp, x, u = integrator_control_problem(stage_method=method, x0 = 0)
        ocp.subject_to(ocp.at_t0(x)==2)   
        with self.assertRaises(Exception):
          sol = ocp.solve()
        if redirect_stdout is not None:
          with StringIO() as buf, redirect_stdout(buf):
            ocp.show_infeasibilities(1e-4)
            out = buf.getvalue()
          self.assertIn("ocp.subject_to(ocp.at_t0(x)==2)",out)
      

    def test_time_dep_ode(self):
        t0 = 1.2
        T = 5.7
        ocp = Ocp(t0=t0,T=5.7)
        
        x = ocp.state()
        ocp.set_der(x, ocp.t**2)

        ocp.subject_to(ocp.at_t0(x)==0)
        
        tf = t0+T
        x_ref = tf**3/3-t0**3/3

        ocp.solver('ipopt')
        opts = {"abstol": 1e-9, "reltol": 1e-9}
        for method in [
                MultipleShooting(intg='rk'),
                MultipleShooting(intg='cvodes',intg_options=opts),
                MultipleShooting(intg='idas',intg_options=opts),
                DirectCollocation()]:
            ocp.method(method)
            sol = ocp.solve()
            ts, xs = sol.sample(x,grid='control')
            x_ref = ts**3/3-t0**3/3
            assert_array_almost_equal(xs,x_ref)
        for method in [SplineMethod(N=4)]:
            ocp.method(method)
            with self.assertRaisesRegex(Exception,"Time dependent variables not supported in SplineMethod"):
              sol = ocp.solve()


    def test_variables(self):
        N = 10
        ocp = Ocp(t0=2*pi,T=10)
        p = ocp.parameter(grid='control')
        v = ocp.variable(grid='control')
        x = ocp.state()
        ocp.set_der(x, 0)
        ocp.subject_to(ocp.at_t0(x)==0)

        ts = linspace(0, 10, N+1)[:-1]

        ocp.add_objective(ocp.integral(sin(v-p)**2,grid='control'))
        ocp.method(MultipleShooting(N=N))
        ocp.solver('ipopt')
        ocp.set_value(p, ts)
        ocp.set_initial(v, ts)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')

        assert_array_almost_equal(xs[:-1], ts)
        ocp.set_initial(v, 0.1+2*pi+ts)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')
        assert_array_almost_equal(xs[:-1], 2*pi+ts)
        ocp.set_initial(v, 0.1+ocp.t)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')
        assert_array_almost_equal(xs[:-1], 2*pi+ts)
        ocp.set_initial(v, 0.1+2*pi)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')
        with self.assertRaises(AssertionError):
          assert_array_almost_equal(xs[:-1], 2*pi+ts)
        #ocp.set_value(p, ocp.t)
        #sol = ocp.solve()
        #_, xs = sol.sample(v, grid='control')
        #assert_array_almost_equal(xs[:-1], linspace(2*pi, 2*pi+1, N))

    def test_variables_plus(self):
        N = 10
        ocp = Ocp(t0=2*pi,T=10)
        p = ocp.parameter(grid='control',include_last=True)
        v = ocp.variable(grid='control',include_last=True)
        x = ocp.state()
        ocp.set_der(x, 0)
        ocp.subject_to(ocp.at_t0(x)==0)

        ts = linspace(0, 10, N+1)

        ocp.add_objective(ocp.sum(sin(v-p)**2,include_last=True))
        ocp.method(MultipleShooting(N=N))
        ocp.solver('ipopt')
        ocp.set_value(p, ts)
        ocp.set_initial(v, ts)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')

        assert_array_almost_equal(xs, ts)
        ocp.set_initial(v, 0.1+2*pi+ts)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')
        assert_array_almost_equal(xs, 2*pi+ts)
        ocp.set_initial(v, 0.1+ocp.t)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')
        assert_array_almost_equal(xs, 2*pi+ts)
        ocp.set_initial(v, 0.1+2*pi)
        sol = ocp.solve()
        _, xs = sol.sample(v, grid='control')
        with self.assertRaises(AssertionError):
          assert_array_almost_equal(xs, 2*pi+ts)

    def test_integral(self):
        t0 = 1.2
        T = 5.7
        ocp = Ocp(t0=t0,T=T)
        
        x = ocp.state()
        u = ocp.control()
        ocp.set_der(x, u)

        ocp.subject_to(ocp.at_t0(x)==0)
        ocp.subject_to(u<=1)
        f = ocp.integral(x*ocp.t)
        ocp.add_objective(-f) # (t-t0)*t -> t^3/3-t^2/2*t0
        ocp.solver('ipopt')
        opts = {"abstol": 1e-8, "reltol": 1e-8, "quad_err_con": True}
        for method in [
                MultipleShooting(intg='rk'),
                MultipleShooting(intg='cvodes',intg_options=opts),
                #MultipleShooting(intg='idas',intg_options=opts),
                DirectCollocation(),
                #SplineMethod()
                ]:
            ocp.method(method)
            sol = ocp.solve()
            ts, xs = sol.sample(f,grid='control')
            I = lambda t: t**3/3-t**2/2*t0
            x_ref = I(t0+T)-I(t0)

            assert_array_almost_equal(xs[-1],x_ref)

    def test_quad(self):
    
    
        for M in [1,2]:
            methods = [SingleShooting(N=3,M=M),
                           MultipleShooting(N=3,M=M),
                           DirectCollocation(N=3,M=M)]
        
            for method in methods:
                ocp = Ocp(T=6)
                
                u = ocp.control()
                q = ocp.state(quad=True)
                q2 = ocp.state(quad=True)
                ocp.set_der(q, u)
                ocp.set_der(q2, u**2)

                ocp.add_objective(ocp.integral(sumsqr(u-sin(ocp.t)))+ocp.at_tf(q2)**2)
                ocp.subject_to(ocp.at_tf(q)==2)
                ocp.solver('ipopt')
                ocp.method(method)
                sol = ocp.solve()
                
                for kwargs in [{"grid":"control"},{"grid":"integrator"},{"grid":"integrator","refine":4}]:
                    refine = kwargs["refine"] if "refine" in kwargs else None
                    
                    if isinstance(method,DirectCollocation):
                        continue
          
                    [_,usol] = sol.sample(u,**kwargs)
                    [_,qsol] = sol.sample(q,**kwargs)
                    
                    q_ref = np.array(vertcat(0,2*np.cumsum(usol[:-1]))).squeeze()
                    if refine is not None:
                        q_ref = q_ref/refine
                    if kwargs["grid"]=="integrator":
                        q_ref = q_ref/M

                    assert_array_almost_equal(qsol,q_ref)

    def test_collocation_equivalence(self):

      for problem in [vdp, vdp_dae]:

        ocp, x1, x2, u = problem(MultipleShooting(N=6,intg='collocation',intg_options={"number_of_finite_elements": 1, "interpolation_order": 4}))
        ocp.solver("ipopt",{"ipopt.tol":1e-12})
        sol = ocp.solve()

        x1_a = sol.sample(x1, grid='control')[1]
        x2_a = sol.sample(x2, grid='control')[1]
        u_a = sol.sample(u, grid='control')[1]

        ocp, x1, x2, u = problem(DirectCollocation(N=6))
        ocp.solver("ipopt",{"ipopt.tol":1e-12})
        sol = ocp.solve()

        x1_b = sol.sample(x1, grid='control')[1]
        x2_b = sol.sample(x2, grid='control')[1]
        u_b = sol.sample(u, grid='control')[1]

        assert_array_almost_equal(x1_a,x1_b,decimal=12)
        assert_array_almost_equal(x2_a,x2_b,decimal=12)
        assert_array_almost_equal(u_a,u_b,decimal=12)


    def test_dae_methods(self):
     
        ocp, x1, x2, u = vdp_dae(DirectCollocation(N=6,M=1))
        ocp.solver("ipopt")
        sol = ocp.solve()

        uref = sol.sample(u, grid='control')[1]

        zsol = sol.sample(ocp.z, grid='control')[1]
        zref = sol.sample(1 - x2**2, grid='control')[1]
        assert_array_almost_equal(zsol,zref,decimal=2)

        for method in [SingleShooting(N=6,intg='idas'),
                       MultipleShooting(N=6,intg='idas'),
                       MultipleShooting(N=6,intg='collocation')]:
           ocp, x1, x2, u = vdp_dae(method)
           ocp.solver("ipopt")
           sol = ocp.solve()

           usol = sol.sample(u, grid='control')[1]
           zsol = sol.sample(ocp.z, grid='control')[1]
           assert_array_almost_equal(zsol[1:],zref[1:],decimal=2)
           assert_array_almost_equal(uref,usol,decimal=3)

        # Constraints with algebraic variables
        ocp, x1, x2, u = vdp_dae(DirectCollocation(N=6,M=1))

        ocp.subject_to(ocp.at_tf(ocp.z) <= 0.8)
        ocp.solver("ipopt")
        sol = ocp.solve()

        zref = sol.sample(ocp.z, grid='control')[1]
        assert_array_almost_equal(zref[-1],0.8,decimal=6)
        uref = sol.sample(u, grid='control')[1]

        for method in [#SingleShooting(N=6,intg='idas'),
                       MultipleShooting(N=6,intg='idas'),
                       MultipleShooting(N=6,intg='collocation')]:
           ocp, x1, x2, u = vdp_dae(method)

           ocp.subject_to(ocp.at_tf(ocp.z) <= 0.8)
           ocp.solver("ipopt")
           sol = ocp.solve()

           usol = sol.sample(u, grid='control')[1]
           zsol = sol.sample(ocp.z, grid='control')[1]
           assert_array_almost_equal(zsol[1:],zref[1:],decimal=2)
           assert_array_almost_equal(uref,usol,decimal=3)

        
        # Path constraints with algebraic variables
        ocp, x1, x2, u = vdp_dae(DirectCollocation(N=6,M=1),x1limit=False)

        ocp.subject_to(0 <= (ocp.z <= 0.8))
        ocp.solver("ipopt")
        sol = ocp.solve()
        ts,zref = sol.sample(ocp.z, grid='control')
        print(zref,ts/10+0.1)
        uref = sol.sample(u, grid='control')[1]

        for method in [SingleShooting(N=6,intg='idas'),
                       MultipleShooting(N=6,intg='idas')]:
                       #MultipleShooting(N=6,intg='collocation')]:
           ocp, x1, x2, u = vdp_dae(method,x1limit=False)

           ocp.subject_to(0 <= (ocp.z <= 0.8))
           ocp.solver("ipopt",{"ipopt.tol": 1e-5})
           sol = ocp.solve()

           usol = sol.sample(u, grid='control')[1]
           zsol = sol.sample(ocp.z, grid='control')[1]
           #assert_array_almost_equal(zsol[1:],zref[1:],decimal=2)
           #assert_array_almost_equal(uref,usol,decimal=3)

           assert_array_almost_equal(np.max(zsol),0.8,decimal=3)
     
    def test_grid_inf_subject_to(self):
        ocp, x1, x2, u = vdp(MultipleShooting(N=10))
        sol = ocp.solve()
        x1sol = sol.sample(x1, grid='integrator',refine=100)[1]
        self.assertFalse(np.all(x1sol>-0.25))

        for method in  [MultipleShooting, DirectCollocation]:
            margins = [np.inf]
            for M in [1,2,4]:
                ocp, x1, x2, u = vdp(method(N=10,M=M),grid='inf')
                sol = ocp.solve()
                x1sol = sol.sample(x1, grid='integrator',refine=100)[1]
                margin = np.min(x1sol-(-0.25))
                self.assertTrue(np.all(margin>0))
                self.assertTrue(np.all(margin<0.01))

                # Assert that margin shrinks for increasing M
                self.assertTrue(margin<0.5*margins[-1])
                margins.append(margin)
                self.assertTrue(margin)

    def test_grid_integrator_subject_to(self):
        ocp, x1, x2, u = vdp(MultipleShooting(N=10))
        sol = ocp.solve()
        x1sol = sol.sample(x1, grid='integrator',refine=100)[1]
        self.assertFalse(np.all(x1sol>-0.25))

        for method, grids in  [(MultipleShooting,['integrator']), (DirectCollocation,['integrator','integrator_roots'])]:
            for grid in grids:
              margins = [np.inf]
              for M in [1,2,4]:
                  ocp, x1, x2, u = vdp(method(N=10,M=M),grid=grid)
                  sol = ocp.solve()
                  x1sol = sol.sample(x1, grid=grid)[1]
                  x1solf = sol.sample(x1, grid='integrator',refine=100)[1]
                  margin = np.min(x1sol-(-0.25))
                  marginf = np.min(x1solf-(-0.25))
                  self.assertTrue(marginf<1e-5)
                  assert_array_almost_equal(margin, 0, decimal=8)

    def test_dae_casadi(self):
        # cross check with dae_colloation

        xref = 0.1 # chariot reference

        l = 1. #- -> crane, + -> pendulum
        m = 1.
        M = 1.
        g = 9.81

        ocp = Ocp(T=5)

        x = ocp.state()
        y = ocp.state()
        w = ocp.state()
        dx = ocp.state()
        dy = ocp.state()
        dw = ocp.state()

        xa = ocp.algebraic()
        u = ocp.control()


        ocp.set_der(x, dx)
        ocp.set_der(y, dy)
        ocp.set_der(w, dw)
        ddx = (w-x)*xa/m
        ddy = g-y*xa/m
        ddw = ((x-w)*xa - u)/M
        ocp.set_der(dx, ddx)
        ocp.set_der(dy, ddy)
        ocp.set_der(dw, ddw)
        ocp.add_alg((x-w)*(ddx - ddw) + y*ddy + dy*dy + (dx-dw)**2)

        ocp.add_objective(ocp.at_tf((x-xref)*(x-xref) + (w-xref)*(w-xref) + dx*dx + dy*dy))
        ocp.add_objective(ocp.integral((x-xref)*(x-xref) + (w-xref)*(w-xref)))

        ocp.subject_to(-2 <= (u <= 2))

        ocp.subject_to(ocp.at_t0(x)==0)
        ocp.subject_to(ocp.at_t0(y)==l)
        ocp.subject_to(ocp.at_t0(w)==0)
        ocp.subject_to(ocp.at_t0(dx)==0)
        ocp.subject_to(ocp.at_t0(dy)==0)
        ocp.subject_to(ocp.at_t0(dw)==0)
        ocp.subject_to(xa>=0,grid='integrator_roots')

        ocp.set_initial(y, l)
        ocp.set_initial(xa, 9.81)

        # Pick an NLP solver backend
        # NOTE: default scaling strategy of MUMPS leads to a singular matrix error 
        ocp.solver('ipopt',{"ipopt.linear_solver": "mumps","ipopt.mumps_scaling":0,"ipopt.tol":1e-12} )

        # Pick a solution method
        method = DirectCollocation(N=50)
        ocp.method(method)

        # Solve
        sol = ocp.solve()

        assert_array_almost_equal(sol.sample(xa, grid='integrator',refine=1)[1][0], 9.81011622448889)
        assert_array_almost_equal(sol.sample(xa, grid='integrator',refine=1)[1][1], 9.865726317147214)
        assert_array_almost_equal(sol.sample(xa, grid='integrator')[1][0], 9.81011622448889)
        assert_array_almost_equal(sol.sample(xa, grid='integrator')[1][1], 9.865726317147214)
        assert_array_almost_equal(sol.sample(xa, grid='control')[1][0], 9.81011622448889)
        assert_array_almost_equal(sol.sample(xa, grid='control')[1][1], 9.865726317147214)

    def test_to_function(self):
        ocp = Ocp(T=FreeTime(1))

        p = ocp.state()
        v = ocp.state()
        u = ocp.control()
        u2 = ocp.control()

        ocp.set_der(p, v)
        ocp.set_der(v, u+u2)

        pp = ocp.parameter()
        ocp.subject_to(u <= p)
        ocp.subject_to(-1 <= u)
        ocp.subject_to(u2 <= p)
        ocp.subject_to(-1 <= u2)

        ocp.add_objective(ocp.T)
        ocp.subject_to(ocp.at_t0(p) == 0)
        ocp.subject_to(ocp.at_t0(v) == 0)
        ocp.subject_to(ocp.at_tf(p) == 1)
        ocp.subject_to(ocp.at_tf(v) == 0)

        ocp.solver('ipopt')

        N = 6
        ocp.method(MultipleShooting(N=N))

        ocp.set_value(pp, 1)
        sol = ocp.solve()

        f = ocp.to_function('f',[ocp.value(pp),ocp.sample(p,grid='control')[1], ocp.sample(v,grid='control')[1], ocp.sample(u,grid='control-')[1]],[ocp.sample(p,grid='control')[1]])

        print(f(1,1,2,3))

    def test_to_function_init(self):
        N = 10
        method = DirectCollocation(N=N,M=2,scheme="legendre")
        ocp, x1, x2, u = vdp_dae(method)
        alg = ocp.z


        ocp.solver("ipopt",{"ipopt.tol":1e-12,"dump_in":True})
      
 
        DM.rng(1)
        
        alg0 = np.round(DM.rand(N+1,1),decimals=3)
        x0 = np.round(DM.rand(2,N+1),decimals=3)


        ocp.set_initial(alg, MX(alg0)[ocp.t])
        ocp.set_initial(ocp.x, MX(x0)[:,ca.floor(ocp.t)])

        sol = ocp.solve()
        x0_ref = DM.from_file("solver.000000.in.x0.mtx")

        ocp, x1, x2, u = vdp_dae(method)
        alg = ocp.z


        ocp.solver("ipopt",{"ipopt.tol":1e-12,"dump_in":True})
    
        _,states = ocp.sample(ocp.x,grid='control')

        f = ocp.to_function('F',[states,"z"],[states])
        f(x0,alg0)
      
        x0 = DM.from_file("solver.000000.in.x0.mtx")


        assert_array_almost_equal(x0,x0_ref)




    def test_no_solver(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      p = ocp.parameter()
      ocp.set_der(x, u*p)
      ocp.method(MultipleShooting(N=3))
      with self.assertRaisesRegex(Exception, "You forgot to declare a solver"):
        ocp.solve()

    def test_no_method(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      p = ocp.parameter()
      ocp.set_der(x, u*p)
      ocp.solver('ipopt')
      with self.assertRaisesRegex(Exception, "You forgot to declare a method"):
        ocp.solve()

    def test_variable_only(self):
      ocp = Ocp()
      v = ocp.variable()
      ocp.subject_to(v>=5)
      ocp.add_objective(v**2)
      ocp.solver("ipopt")
      sol = ocp.solve()
      self.assertAlmostEqual(sol.value(v),5)

      ocp = Ocp()
      v = ocp.variable()
      ocp.add_objective((v-3)**2)
      ocp.solver("ipopt")
      sol = ocp.solve()
      self.assertAlmostEqual(sol.value(v),3)

      ocp = Ocp()
      v = ocp.variable()
      p = ocp.parameter()
      ocp.add_objective((v-p)**2)
      ocp.solver("ipopt")
      ocp.set_value(p,3)
      sol = ocp.solve()
      ocp.set_value(p,4)
      sol = ocp.solve()
      self.assertAlmostEqual(sol.value(v),4)

      ocp = Ocp()
      v = ocp.variable()
      p = ocp.parameter()
      ocp.subject_to(v>=p)
      ocp.add_objective(v**2)
      ocp.solver("ipopt")
      ocp.set_value(p,5)
      sol = ocp.solve()
      self.assertAlmostEqual(sol.value(v),5)

      ocp.save('test.rockit')

      ocp2 = Ocp.load("test.rockit")
      sol2 = ocp2.solve()
      self.assertAlmostEqual(sol2.value(ocp2.v),5)

    def test_no_objective(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      ocp.set_der(x, u)
      ocp.solver('ipopt')
      ocp.method(MultipleShooting(N=3))
      ocp.solve()

    def test_no_solve(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      ocp.set_der(x, u)
      ocp.solver('ipopt')
      ocp.method(MultipleShooting(N=3))
      with self.assertRaisesRegex(Exception, "You forgot to solve first"):
        ocp.non_converged_solution

    def test_no_set_value(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      p = ocp.parameter()
      ocp.set_der(x, u*p)
      ocp.solver('ipopt')
      ocp.method(MultipleShooting(N=3))
      with self.assertRaisesRegex(Exception, "You forgot to declare a value"):
        ocp.solve()

    def test_set_value_error(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      p = ocp.parameter()
      ocp.set_der(x, u*p)
      ocp.solver('ipopt')
      ocp.method(MultipleShooting(N=3))

      y = MX.sym('y')

      ocp.set_value(p, 2)
      with self.assertRaisesRegex(Exception, "You attempted to set the value of a non-parameter"):
        ocp.set_value(y, 3)
      with self.assertRaisesRegex(Exception, "You attempted to set the value of a non-parameter"):
        ocp.set_value(x, 3)


      ocp.solve()
      ocp.set_value(p, 2)
      with self.assertRaisesRegex(Exception, "You attempted to set the value of a non-parameter"):
        ocp.set_value(x, 3)
      with self.assertRaisesRegex(Exception, "You attempted to set the value of a non-parameter"):
        ocp.set_value(y, 3)

    def test_set_initial(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      p = ocp.parameter()
      y = MX.sym('y')
      ocp.set_der(x, u*p)
      ocp.solver('ipopt')
      ocp.method(MultipleShooting(N=3))
      ocp.set_value(p, 2)

      ocp.set_initial(x, 2)
      with self.assertRaisesRegex(Exception, "You attempted to set the initial value of a parameter"):
        ocp.set_initial(p, 3)
      with self.assertRaisesRegex(Exception, "You attempted to set the initial value of an unknown symbol"):
        ocp.set_initial(y, 3)
      ocp.solve()
      ocp.set_initial(x, 2)
      with self.assertRaisesRegex(Exception, "You attempted to set the initial value of a parameter"):
        ocp.set_initial(p, 3)
      with self.assertRaisesRegex(Exception, "You attempted to set the initial value of an unknown symbol"):
        ocp.set_initial(y, 3)

    def test_set_initial_alg(self):
    
      for method in [SingleShooting(N=10,intg='idas'),MultipleShooting(N=10,intg='idas'),DirectCollocation(N=10)]:
        ocp = Ocp()
        x = ocp.state()
        zx = ocp.algebraic()
        y = ocp.state()
        zy = ocp.algebraic()
        
        ocp.add_alg(sin(zx))
        ocp.add_alg(ca.cos(zy))
        
        u = ocp.control()
        
        ocp.set_der(x, u/zx)
        ocp.set_der(y, u/zy)
        
        ocp.subject_to(ocp.at_t0(x)==0)
        ocp.subject_to(ocp.at_t0(y)==0)
        
        ocp.add_objective(ocp.sum(u**2))
        
        ocp.subject_to(ocp.at_tf(x)==1)
        
        ocp.method(method)
        ocp.solver("ipopt")
        
        with self.assertRaises(Exception):
        	sol = ocp.solve()
        
        ocp.set_initial(zy, pi/2-0.1)
        ocp.set_initial(zx, pi-0.1)
        
        sol = ocp.solve()
        
        np.testing.assert_allclose(sol.sample(zy,grid='control')[1],pi/2)
        np.testing.assert_allclose(sol.sample(zx,grid='control')[1],pi)
      
      for method in [MultipleShooting(N=4,intg='idas'),SingleShooting(N=4,intg='idas'),DirectCollocation(N=4,scheme='legendre')]:
        ocp = Ocp(T=4)
        x = ocp.state()
        zx = ocp.algebraic()
        y = ocp.state()
        zy = ocp.algebraic()
        
        ocp.add_alg(sin(zx))
        ocp.add_alg(ca.cos(zy))
        
        ocp.set_initial(zy, pi/2+pi*ca.floor(ocp.t))
        ocp.set_initial(zx, pi+pi*ca.floor(ocp.t))
        
        u = ocp.control()
        
        ocp.set_der(x, u/zx)
        ocp.set_der(y, u/zy)
        
        ocp.subject_to(ocp.at_t0(x)==0)
        ocp.subject_to(ocp.at_t0(y)==0)
        
        ocp.add_objective(ocp.sum(u**2))
        
        ocp.subject_to(ocp.at_tf(x)==1)
        
        ocp.method(method)
        ocp.solver("ipopt")
        
        sol = ocp.solve()
        
        print(sol.sample(zy,grid='control')[1])
        print(sol.sample(zx,grid='control')[1])
        print(pi/2+pi*np.linspace(0,4,4+1))

        if "MultipleShooting" in str(method):
          # Note the first sample of z is based on discrete_system Zi[0] and Zi is the collection of integrator outputs zf.
          # So for M=1, sampled z is same for k=0 and k=1.
          

          # Furthermore, the first zf here gets initial guess (pi/2), leading to pi/2 solution
          np.testing.assert_allclose(sol.sample(zy,grid='control')[1],[pi/2,pi/2,pi/2+pi,pi/2+2*pi,pi/2+3*pi])
          np.testing.assert_allclose(sol.sample(zx,grid='control')[1],[pi,pi,pi+pi,pi+2*pi,pi+3*pi])
        elif "SingleShooting" in str(method):
          # In single shooting, only the initial guess (pi/2) is used
          np.testing.assert_allclose(sol.sample(zy,grid='control')[1],pi/2)
          np.testing.assert_allclose(sol.sample(zx,grid='control')[1],pi)
        elif "DirectCollocation" in str(method):
          legendre = [pi/2]*4+[pi/2+pi]*4+[pi/2+2*pi]*4+[pi/2+3*pi]*4
          np.testing.assert_allclose(sol.sample(zy,grid='integrator_roots')[1],legendre)
          # Note sure if the answer below is desirable
          np.testing.assert_allclose(sol.sample(zy,grid='control')[1],[pi/2,pi/2+pi,pi/2+2*pi,pi/2+3*pi,pi/2+3*pi])

      
    def test_control_set_der(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      with self.assertRaisesRegex(Exception, "You used set_der on a non"):
        ocp.set_der(u, 2)


    def test_add_objective_signal(self):
      ocp = Ocp()
      x = ocp.state()
      u = ocp.control()
      p = ocp.parameter()
      y = MX.sym('y')
      ocp.set_der(x, u*p)
      ocp.solver('ipopt')
      with self.assertRaisesRegex(Exception, "An objective cannot be a signal"):
        ocp.add_objective(x)

    def test_localize_time(self):
      N = 10
      for t0_stage in [FreeTime(-1), -1]:
          for T_stage in [FreeTime(2), 2]:
              t0_free = isinstance(t0_stage, FreeTime)
              T_free = isinstance(T_stage, FreeTime)
              ocp = Ocp(t0=t0_stage,T=T_stage)

              p = ocp.state()
              v = ocp.state()
              u = ocp.control()

              ocp.set_der(p, v)
              ocp.set_der(v, u)

              ocp.add_objective(ocp.tf)
              ocp.subject_to(ocp.at_t0(p) == 0)
              ocp.subject_to(ocp.at_t0(v) == 0)
              ocp.subject_to(ocp.at_tf(p) == 1)
              ocp.subject_to(ocp.at_tf(v) == 0)

              ocp.subject_to(u <= 1)
              ocp.subject_to(-1 <= u)
 
              ocp.solver('ipopt',{"ipopt.tol":1e-12})

              if t0_free:
                  ocp.subject_to(ocp.t0 >= -1)

              for localize_t0 in [False,True]:
                if (not t0_free) and localize_t0: continue
                for localize_T in [False,True]:
                  if (not T_free) and localize_T: continue
  
                  ocp.method(MultipleShooting(N=N,grid=UniformGrid(localize_T=localize_T,localize_t0=localize_t0)))

                  sol = ocp.solve()
                  opti = ocp._method.opti
                  self.assertTrue(sum2(sum1(DM(jacobian(opti.g, opti.x).sparsity(),1))>N)<= t0_free+T_free-localize_t0-localize_T)
                    

    def test_control_grid(self):
      N = 4
      # Be careful for loss of symmetry for non-uniform grids; bang-bang is only optimal if normalized(t=0.5) is included
      # e.g. for local growth factor r: (1-r^7)/(1-r^10) = 0.5 -> r = 
      r = MX.sym("r")
      rf = rootfinder("rf","newton",{"x": r, "g": (1-r**3)/(1-r**N)-0.5})
      r = float(rf(x0=1.2)["x"])
      for t0_stage in [FreeTime(-1), -1]:
          for T_stage in [FreeTime(2), 2]:
              t0_free = isinstance(t0_stage, FreeTime)
              T_free = isinstance(T_stage, FreeTime)
              ocp = Ocp(t0=t0_stage,T=T_stage)

              p = ocp.state()
              v = ocp.state()
              u = ocp.control()

              ocp.set_der(p, v)
              ocp.set_der(v, u)

              ocp.add_objective(ocp.tf)
              ocp.subject_to(ocp.at_t0(p) == 0)
              ocp.subject_to(ocp.at_t0(v) == 0)
              ocp.subject_to(ocp.at_tf(p) == 1)
              ocp.subject_to(ocp.at_tf(v) == 0)


              ocp.subject_to(-1 <= (u <= 1))
              ocp.solver('ipopt',{"ipopt.tol":1e-12})

              if t0_free:
                  ocp.subject_to(ocp.t0 >= -1)

              for localize_t0 in [False,True]:
                if (not t0_free) and localize_t0: continue
                for localize_T in [False,True]:
                  if (not T_free) and localize_T: continue

                  for grid in [
                    UniformGrid(localize_T=localize_T,localize_t0=localize_t0),
                    GeometricGrid(r,local=True,localize_T=localize_T,localize_t0=localize_t0),
                    GeometricGrid(r**(N-1),local=False,localize_T=localize_T,localize_t0=localize_t0),
                    FreeGrid(min=0.01,max=0.5)]:
                      ocp.method(MultipleShooting(N=N,grid=grid))

                      sol = ocp.solve()

                      tolerance = 1e-6

                      if isinstance(grid, UniformGrid):

                        ts_ref = -1 + np.linspace(0, 2, 4*N+1)

                        ts, _ = sol.sample(v, grid='integrator', refine=4)
                        np.testing.assert_allclose(ts, ts_ref, atol=tolerance)

                      ts, _ = sol.sample(v, grid='control')
                      if not isinstance(grid, FreeGrid):

                        ts_ref = -1 + 2*np.array(grid.normalized(N))

                        np.testing.assert_allclose(ts, ts_ref, atol=tolerance)

                      np.testing.assert_allclose(ts[0], -1, atol=tolerance)
                      np.testing.assert_allclose(ts[-1], 1, atol=tolerance)

                      self.assertAlmostEqual(sol.value(ocp.t0),-1, places=5)
                      self.assertAlmostEqual(sol.value(ocp.T),2, places=5)
                      self.assertAlmostEqual(sol.value(ocp.tf),1, places=5)

    def test_augmented_set_initial(self):
      ocp = Ocp(t0=FreeTime(0),T=FreeTime(2))

      p = ocp.state()
      v = ocp.state()
      u = ocp.control()

      ocp.set_der(p, v)
      ocp.set_der(v, u)

      ocp.add_objective(ocp.tf)
      ocp.subject_to(ocp.at_t0(p) == 0)
      ocp.subject_to(ocp.at_t0(v) == 0)
      ocp.subject_to(ocp.at_tf(p) == 1)
      ocp.subject_to(ocp.at_tf(v) == 0)
      ocp.subject_to(ocp.t0 == 0)
      ocp.subject_to(ocp.tf == 2)

      ocp.subject_to(-1 <= (u <= 1))
      ocp.solver('ipopt',{"ipopt.tol":1e-12})

      ocp.method(MultipleShooting(N=4))

      sol = ocp.solve()
      ocp.set_initial(ocp.t0, 2)
      ocp.set_initial(ocp.T, 2)
      ocp.solve()
      
      
    def test_der_t(self):
      ocp = Ocp()

      r = evalf(ocp.der(ocp.t))
      self.assertEqual(r,1)

    def test_subject_to_ocp_t(self):
      for method in [MultipleShooting(N=10),SingleShooting(N=10),DirectCollocation(N=10),SplineMethod(N=10)]:
        ocp = Ocp(T=10)
        p = ocp.state()
        v = ocp.state()
        u = ocp.control()
        ocp.set_der(p, v)
        ocp.set_der(v, u)
        ocp.add_objective(-ocp.at_tf(p))

        ocp.subject_to(v<= 2+sin(ocp.t))

        ocp.subject_to(ocp.at_t0(p)==0)

        ocp.method(method)
        ocp.solver('ipopt')

        sol = ocp.solve()

        [ts,vsol] = sol.sample(v,grid='control')
        np.testing.assert_allclose(vsol, 2+sin(ts), atol=1e-6)

      for method in [MultipleShooting(N=10),SingleShooting(N=10),DirectCollocation(N=10),SplineMethod(N=10)]:
        ocp = Ocp(T=FreeTime(10))
        p = ocp.state()
        v = ocp.state()
        u = ocp.control()
        ocp.set_der(p, v)
        ocp.set_der(v, u)
        ocp.add_objective(ocp.T)

        ocp.subject_to(v<= 2+sin(ocp.t))

        ocp.subject_to(ocp.at_t0(p)==0)
        ocp.subject_to(ocp.at_tf(p)==5)

        ocp.method(method)
        ocp.solver('ipopt')

        sol = ocp.solve()

        [ts,vsol] = sol.sample(v,grid='control')
        np.testing.assert_allclose(vsol, 2+sin(ts), atol=1e-6)

    def test_exploiting_constraints(self):
        # Get as far as possible with a speed limit
        # If we are not careful with enforcing the speed limit, we may end up in cheating solutions that oscillate
        for N in [9,10]:
          for method in [SplineMethod(N=N),MultipleShooting(N=N),SingleShooting(N=N),DirectCollocation(N=N),SplineMethod(N=N)]:
            escape = N % 2 == 1

            for refine_subject_to in [True,False]:
              if refine_subject_to and not isinstance(method,SplineMethod): continue
              ocp = Ocp(T=10)
              p = ocp.control(order=3)
              v = ocp.der(p)
              ocp.add_objective(-ocp.at_tf(p))

              ocp.subject_to(v<=3,grid='inf')

              kwargs = {}
              if refine_subject_to:
                kwargs["refine"] = 2

              ocp.subject_to(v<=2, **kwargs)

              ocp.subject_to(ocp.at_t0(p)==0)

              ocp.method(method)
              ocp.solver('ipopt')

              sol = ocp.solve()

              [ts,vsol] = sol.sample(v,grid='control')
              [tsf,vfsol] = sol.sample(v,grid='control' if isinstance(method, SplineMethod) else 'integrator',refine=10)
              if escape and not refine_subject_to:
                print(method)
                np.testing.assert_allclose(vsol, 2, atol=1e-6)
                print(vfsol)
                # Cheats
                self.assertTrue(vfsol[5]>=2.3)
              else:
                np.testing.assert_allclose(vfsol, 2, atol=1e-6)
          #import pylab as plt
          #plt.plot(ts,vsol,'o')
          #plt.plot(tsf,vfsol,'-')
          #plt.show()
          #np.testing.assert_allclose(vsol, 2+sin(ts), atol=1e-6)

    def test_subject_to_group_refine(self):
        ocp = Ocp(T=10)
        p = ocp.control(order=3)
        v = ocp.der(p)
        ocp.add_objective(-ocp.at_tf(p))

        ocp.subject_to(v<=3,grid='inf')

        ocp.subject_to(v<=2)

        ocp.subject_to(ocp.at_t0(p)==0)

        ocp.method(SplineMethod(N=9))
        ocp.solver('ipopt')

        sol = ocp.solve()
        size_origin = ocp.jacobian().size1()

        ocp = Ocp(T=10)
        p = ocp.control(order=3)
        v = ocp.der(p)
        ocp.add_objective(-ocp.at_tf(p))

        ocp.subject_to(v<=3,grid='inf')

        ocp.subject_to(v<=2, refine=2)

        ocp.subject_to(ocp.at_t0(p)==0)

        ocp.method(SplineMethod(N=9))
        ocp.solver('ipopt')

        sol = ocp.solve()

        [ts,vsol] = sol.sample(v,grid='control')
        [tsf,vfsol] = sol.sample(v,grid='control',refine=10)
        size_normal = ocp.jacobian().size1()
        self.assertTrue(size_normal>size_origin)

        ocp = Ocp(T=10)
        p = ocp.control(order=3)
        v = ocp.der(p)
        ocp.add_objective(-ocp.at_tf(p))

        ocp.subject_to(v<=3,grid='inf')

        ocp.subject_to(v<=2, refine=2, group_refine=LseGroup(margin_abs=0.1))

        ocp.subject_to(ocp.at_t0(p)==0)

        ocp.method(SplineMethod(N=9))
        ocp.solver('ipopt')

        sol = ocp.solve()

        [ts,vsol] = sol.sample(v,grid='control')
        [tsf,vfsol] = sol.sample(v,grid='control',refine=10)

        # Check that grouping satisfies the constraint
        np.testing.assert_array_less(vfsol, 2+1e-8)

        size_lse = ocp.jacobian().size1()
        self.assertEqual(size_lse,size_origin)

        # Lower bound active

        ocp = Ocp(T=10)
        p = ocp.control(order=3)
        v = ocp.der(p)
        ocp.add_objective(-ocp.at_tf(p))

        nv = -v

        ocp.subject_to(v<=3,grid='inf')

        ocp.subject_to(-2<=nv, refine=2, group_refine=LseGroup(margin_abs=0.1))

        ocp.subject_to(ocp.at_t0(p)==0)

        ocp.method(SplineMethod(N=9))
        ocp.solver('ipopt')

        sol = ocp.solve()
        [ts,vsol] = sol.sample(v,grid='control')
        [tsf,vfsol2] = sol.sample(v,grid='control',refine=10)

        size_lse2 = ocp.jacobian().size1()
        np.testing.assert_allclose(vfsol, vfsol2, atol=1e-6)
        self.assertEqual(size_lse,size_lse2)


    def test_initial_spline(self):
        for order in range(1,4):
          for power in range(2): # Does not work for 2; set_initial is correctly setting the control points; but a spline does not pass exactly through control points
            t0 = 1
            T = 5
            ocp = Ocp(t0=1,T=T)
            x = ocp.control(order=order)
            v0 = 3
            x0 = 2
            v = ocp.der(x)
            ocp.subject_to(ocp.at_t0(x)==0)
            ocp.set_initial(x, x0 + v0*ocp.t**power)
            ocp.method(SplineMethod(N=4))
            ocp.solver('ipopt',{'ipopt.max_iter':0})
            print(order,power)
            sol = ocp.solve_limited()

            [ts,xsol] = sol.sample(x,grid='control')
            np.testing.assert_allclose(xsol, x0 + v0*ts**power, atol=1e-6)

            [tsf,xfsol] = sol.sample(x,grid='control',refine=10)
            [_,vfsol] = sol.sample(v,grid='control',refine=10)

            np.testing.assert_allclose(xfsol, x0 + v0*tsf**power, atol=1e-6)
            np.testing.assert_allclose(vfsol, power*v0*tsf**(power-1), atol=1e-6)

    def test_initial_localizeT(self):
        for localize_T in [False,True]:
            ocp = Ocp(T=FreeTime(10))
            x = ocp.state()
            u = ocp.control()
            ocp.set_der(x,u)
            ocp.subject_to(ocp.at_t0(x)==0)
            ocp.subject_to(ocp.at_tf(x)==10)
            ocp.subject_to(-1<= (u<= 1))
            ocp.set_initial(x, ocp.t)
            ocp.method(MultipleShooting(N=4,grid=UniformGrid(localize_T=localize_T)))
            ocp.solver('ipopt',{'ipopt.max_iter':0})
            sol = ocp.solve_limited()

            [ts,xsol] = sol.sample(x,grid='control')
            np.testing.assert_allclose(ts, xsol, atol=1e-6)
  
    def test_samplingmethod_architecture(self):
        for method in [MultipleShooting(N=4),SingleShooting(N=4),DirectCollocation(N=4),SplineMethod(N=4)]:
            print(method)
            (ocp, p, v, a) = bang_bang_problem(method)
            alpha = ocp.parameter()
            ocp.set_value(alpha,3)
            ocp.add_objective(alpha*ocp.integral(a**2))
            [_,psol] = ocp.sample(p)
            
            syms = symvar(psol)
            self.assertTrue(np.all(["opti" in e.name() for e in syms]))

            [_,psol] = ocp.sample(alpha*p)
            
            syms = symvar(psol)
            self.assertTrue(np.all(["opti" in e.name() for e in syms]))

    def test_pickling(self):
      for method in [MultipleShooting(N=4),SingleShooting(N=4),DirectCollocation(N=4),SplineMethod(N=4)]:
        (ocp, p, v, a) = bang_bang_problem(method)

        sol = ocp.solve()
        [tr,xr] = sol.sample(ocp.x,grid='control')
        ocp.save("foo.rockit")
        ocp2 = Ocp.load("foo.rockit")

        sol2 = ocp2.solve()
        [t2,x2] = sol2.sample(ocp2.x,grid='control')

        np.testing.assert_allclose(tr, t2,atol=1e-16)
        np.testing.assert_allclose(xr, x2,atol=1e-16)

    def test_set_initial_bspline_variable(self):

        ocp = Ocp(T=10)

        def R_compress(R):
            return ca.vec(R)


        def R_decompress(Rc):
            return ca.reshape(Rc,3,3)


        Rc = ocp.variable(9, grid='bspline', order=4)
        R = R_decompress(Rc)


        ocp.subject_to(ocp.at_t0(R)==ca.DM.eye(3))

        #ocp.subject_to(ca.vec(R.T @ R-ca.DM.eye(3))==0)

        ocp.method(SplineMethod(N=5))

        ocp.solver("ipopt",{"ipopt.max_iter":0})

        print(ca.densify(R_compress(ca.DM.eye(3))))
        ocp.set_initial(Rc,ca.densify(R_compress(ca.DM.eye(3))))

        sol = ocp.solve_limited()

        for Rc_sol in ca.vertsplit(sol.sample(Rc,grid='control')[1]):
          assert_array_almost_equal(R_decompress(Rc_sol),DM.eye(3))

    def test_variable_bspline(self):
        
        N = 3
        M = 2
        
        for grid in [UniformGrid(),  GeometricGrid(2)]:
          for degree in [1,2,0]:
            ocp = Ocp(T=10)

            # Define 2 states
            x1 = ocp.state()
            x2 = ocp.state()

            # Define 1 control
            u = ocp.control(order=degree)

            # Specify ODE
            ocp.set_der(x1, (1 - x2**2) * x1 - x2 + u)
            ocp.set_der(x2, x1)

            # Lagrange objective
            ocp.add_objective(ocp.integral(x1**2 + x2**2 + u**2))

            # Path constraints
            ocp.subject_to(-1 <= (u <= 1))
            ocp.subject_to(x1 >= -0.25, grid='integrator')

            # Initial constraints
            ocp.subject_to(ocp.at_t0(x1) == 0)
            ocp.subject_to(ocp.at_t0(x2) == 1)

            # Pick an NLP solver backend
            ocp.solver('ipopt')

            # Pick a solution method
            ocp.method(DirectCollocation(N=N,M=M,grid=grid))
            
            sol = ocp.solve()

            [_,urefsol_control] = sol.sample(u,grid='control')
            [_,urefsol_integrator] = sol.sample(u,grid='integrator')
            [_,urefsol_integrator_refine] = sol.sample(u,grid='integrator',refine=7)
            [_,urefsol_integrator_roots] = sol.sample(u,grid='integrator_roots')


            ocp = Ocp(T=10)

            # Define 2 states
            x1 = ocp.state()
            x2 = ocp.state()

            # Define 1 control
            u = ocp.variable(grid='bspline',order=degree)

            # Specify ODE
            ocp.set_der(x1, (1 - x2**2) * x1 - x2 + u)
            ocp.set_der(x2, x1)

            # Lagrange objective
            ocp.add_objective(ocp.integral(x1**2 + x2**2 + u**2))

            # Path constraints
            ocp.subject_to(-1 <= (u <= 1))
            ocp.subject_to(x1 >= -0.25, grid='integrator')

            # Initial constraints
            ocp.subject_to(ocp.at_t0(x1) == 0)
            ocp.subject_to(ocp.at_t0(x2) == 1)

            # Pick an NLP solver backend
            ocp.solver('ipopt')

            # Pick a solution method
            ocp.method(DirectCollocation(N=N,M=M,grid=grid))
            
            sol = ocp.solve()

            [_,usol_control] = sol.sample(u,grid='control')
            #[_,usol_integrator] = sol.sample(u,grid='integrator')
            [_,usol_integrator_refine] = sol.sample(u,grid='integrator',refine=7)
            #[_,usol_integrator_roots] = sol.sample(u,grid='integrator_roots')

            assert_array_almost_equal(urefsol_control,usol_control)
            #assert_array_almost_equal(urefsol_integrator,usol_integrator)
            assert_array_almost_equal(urefsol_integrator_refine,usol_integrator_refine)
            #assert_array_almost_equal(urefsol_integrator_roots,usol_integrator_roots)
            # Set initial



if __name__ == '__main__':
    unittest.main()





# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#
# cython: language_level=3
# cython: profile=False
# distutils: language=c

cimport cython
from libc cimport string
# from libc cimport bool as bool_t

cimport acados_sim_solver_common
cimport acados_sim_solver

cimport numpy as cnp

import os
from datetime import datetime
import numpy as np


cdef class AcadosSimSolverCython:
    """
    Class to interact with the acados sim solver C object.
    """

    cdef acados_sim_solver.sim_solver_capsule *capsule
    cdef void *sim_dims
    cdef acados_sim_solver_common.sim_opts *sim_opts
    cdef acados_sim_solver_common.sim_config *sim_config
    cdef acados_sim_solver_common.sim_out *sim_out
    cdef acados_sim_solver_common.sim_in *sim_in
    cdef acados_sim_solver_common.sim_solver *sim_solver

    cdef bint solver_created

    cdef str model_name

    cdef str sim_solver_type

    cdef list gettable_vectors
    cdef list gettable_matrices
    cdef list gettable_scalars

    def __cinit__(self, model_name):

        self.solver_created = False

        self.model_name = model_name

        # create capsule
        self.capsule = acados_sim_solver.acados_sim_solver_create_capsule()

        # create solver
        assert acados_sim_solver.acados_sim_create(self.capsule) == 0
        self.solver_created = True

        # get pointers solver
        self.__get_pointers_solver()

        self.gettable_vectors = ['x', 'u', 'z', 'S_adj']
        self.gettable_matrices = ['S_forw', 'Sx', 'Su', 'S_hess', 'S_algebraic']
        self.gettable_scalars = ['CPUtime', 'time_tot', 'ADtime', 'time_ad', 'LAtime', 'time_la']

    def __get_pointers_solver(self):
        """
        Private function to get the pointers for solver
        """
        # get pointers solver
        self.sim_opts = acados_sim_solver.acados_get_sim_opts(self.capsule)
        self.sim_dims = acados_sim_solver.acados_get_sim_dims(self.capsule)
        self.sim_config = acados_sim_solver.acados_get_sim_config(self.capsule)
        self.sim_out = acados_sim_solver.acados_get_sim_out(self.capsule)
        self.sim_in = acados_sim_solver.acados_get_sim_in(self.capsule)
        self.sim_solver = acados_sim_solver.acados_get_sim_solver(self.capsule)


    def simulate(self, x=None, u=None, z=None, p=None):
        """
        Simulate the system forward for the given x, u, z, p and return x_next.
        Wrapper around `solve()` taking care of setting/getting inputs/outputs.
        """
        if x is not None:
            self.set('x', x)
        if u is not None:
            self.set('u', u)
        if z is not None:
            self.set('z', z)
        if p is not None:
            self.set('p', p)

        status = self.solve()

        if status == 2:
            print("Warning: acados_sim_solver reached maximum iterations.")
        elif status != 0:
            raise Exception(f'acados_sim_solver for model {self.model_name} returned status {status}.')

        x_next = self.get('x')
        return x_next


    def solve(self):
        """
        Solve the sim with current input.
        """
        return acados_sim_solver.acados_sim_solve(self.capsule)


    def get(self, field_):
        """
        Get the last solution of the solver.

            :param str field: string in ['x', 'u', 'z', 'S_forw', 'Sx', 'Su', 'S_adj', 'S_hess', 'S_algebraic', 'CPUtime', 'time_tot', 'ADtime', 'time_ad', 'LAtime', 'time_la']
        """
        field = field_.encode('utf-8')

        if field_ in self.gettable_vectors:
            return self.__get_vector(field)
        elif field_ in self.gettable_matrices:
            return self.__get_matrix(field)
        elif field_ in self.gettable_scalars:
            return self.__get_scalar(field)
        else:
            raise Exception(f'AcadosSimSolver.get(): Unknown field {field_},' \
                f' available fields are {", ".join(self.gettable.keys())}')


    def __get_scalar(self, field):
        cdef double scalar
        acados_sim_solver_common.sim_out_get(self.sim_config, self.sim_dims, self.sim_out, field, <void *> &scalar)
        return scalar


    def __get_vector(self, field):
        cdef int[2] dims
        acados_sim_solver_common.sim_dims_get_from_attr(self.sim_config, self.sim_dims, field, &dims[0])
        # cdef cnp.ndarray[cnp.float64_t, ndim=1] out = np.ascontiguousarray(np.zeros((dims[0],), dtype=np.float64))
        cdef cnp.ndarray[cnp.float64_t, ndim=1] out = np.zeros((dims[0]),)
        acados_sim_solver_common.sim_out_get(self.sim_config, self.sim_dims, self.sim_out, field, <void *> out.data)
        return out


    def __get_matrix(self, field):
        cdef int[2] dims
        acados_sim_solver_common.sim_dims_get_from_attr(self.sim_config, self.sim_dims, field, &dims[0])
        cdef cnp.ndarray[cnp.float64_t, ndim=2] out = np.zeros((dims[0], dims[1]), order='F', dtype=np.float64)
        acados_sim_solver_common.sim_out_get(self.sim_config, self.sim_dims, self.sim_out, field, <void *> out.data)
        return out


    def set(self, field_: str, value_):
        """
        Set numerical data inside the solver.

            :param field: string in ['p', 'seed_adj', 'T', 'x', 'u', 'xdot', 'z']
            :param value: the value with appropriate size.
        """
        settable = ['seed_adj', 'T', 'x', 'u', 'xdot', 'z', 'p'] # S_forw

        # cast value_ to avoid conversion issues
        if isinstance(value_, (float, int)):
            value_ = np.array([value_])
        # if len(value_.shape) > 1:
            # raise RuntimeError('AcadosSimSolverCython.set(): value_ should be 1 dimensional')

        cdef cnp.ndarray[cnp.float64_t, ndim=1] value = np.ascontiguousarray(value_, dtype=np.float64).flatten()

        field = field_.encode('utf-8')
        cdef int[2] dims

        # treat parameters separately
        if field_ == 'p':
            assert acados_sim_solver.acados_sim_update_params(self.capsule, <double *> value.data, value.shape[0]) == 0
            return
        else:
            acados_sim_solver_common.sim_dims_get_from_attr(self.sim_config, self.sim_dims, field, &dims[0])

            value_ = np.ravel(value_, order='F')

            value_shape = value_.shape
            if len(value_shape) == 1:
                value_shape = (value_shape[0], 0)

            if value_shape != tuple(dims):
                raise Exception(f'AcadosSimSolverCython.set(): mismatching dimension' \
                    f' for field "{field_}" with dimension {tuple(dims)} (you have {value_shape}).')

        # set
        if field_ in ['xdot', 'z']:
            acados_sim_solver_common.sim_solver_set(self.sim_solver, field, <void *> value.data)
        elif field_ in settable:
            acados_sim_solver_common.sim_in_set(self.sim_config, self.sim_dims, self.sim_in, field, <void *> value.data)
        else:
            raise Exception(f'AcadosSimSolverCython.set(): Unknown field {field_},' \
                f' available fields are {", ".join(settable)}')


    def options_set(self, field_: str, value_: bool):
        """
        Set solver options

            :param field: string in ['sens_forw', 'sens_adj', 'sens_hess']
            :param value: Boolean
        """
        fields = ['sens_forw', 'sens_adj', 'sens_hess']
        if field_ not in fields:
            raise Exception(f"field {field_} not supported. Supported values are {', '.join(fields)}.\n")

        field = field_.encode('utf-8')

        if not isinstance(value_, bool):
            raise TypeError("options_set: expected boolean for value")

        cdef bint bool_value = value_
        acados_sim_solver_common.sim_opts_set(self.sim_config, self.sim_opts, field, <void *> &bool_value)
        # TODO: only allow setting
        # if getattr(self.acados_sim.solver_options, field_) or value_ == False:
        #     acados_sim_solver_common.sim_opts_set(self.sim_config, self.sim_opts, field, <void *> &bool_value)
        # else:
        #     raise RuntimeError(f"Cannot set option {field_} to True, because it was False in original solver options.\n")

        return


    def __del__(self):
        if self.solver_created:
            acados_sim_solver.acados_sim_free(self.capsule)
            acados_sim_solver.acados_sim_solver_free_capsule(self.capsule)



import unittest

# Make available sin, cos, etc
from numpy import *
# Import the project
from rockit import *
import casadi as cs
import glob
import os

intg_options = {"rootfinder": "nlpsol", "rootfinder_options": {"nlpsol":"ipopt","nlpsol_options":{"ipopt.print_level":0,"print_time":False,"ipopt.tol":1e-10}}}

class ScalingTests(unittest.TestCase):


    def test_vdp(self):

        for Method in [lambda **kwargs: SingleShooting(intg='collocation',intg_options=intg_options,**kwargs), lambda **kwargs: MultipleShooting(intg='collocation',intg_options=intg_options,**kwargs), lambda **kwargs:  DirectCollocation(**kwargs)]:

            data_ref = None
            for run_type in ["ref","scaled"]:
                #%%
                # Problem specification
                # ---------------------

                # Start an optimal control environment with a time horizon of 10 seconds
                # starting from t0=0s.
                #  (free-time problems can be configured with `FreeTime(initial_guess)`)
                ocp = Ocp(t0=0, T=10)

                s1 = 1
                s2 = 1
                su = 1
                s = 1
                si = 1
                sz = 1
                
                if run_type=="scaled":
                    s1 = 2
                    s2 = 3
                    su = 5
                    s = 7
                    si = 11
                    sz = 13

                # Define two scalar states (vectors and matrices also supported)
                x1 = ocp.state(scale=1/s1) # [m/s]
                x2 = ocp.state(scale=1/s2) # [um/s]

                # Define one piecewise constant control input
                #  (use `order=1` for piecewise linear)
                u = ocp.control(scale=1/su)

                z = ocp.algebraic(scale=1/sz)
                
                # Compose time-dependent expressions a.k.a. signals
                #  (explicit time-dependence is supported with `ocp.t`)
                ocp.add_alg(sz*z-(1 - 1e-2*(s2*x2)**2),scale=sz)

                # Specify differential equations for states
                #  (DAEs also supported with `ocp.algebraic` and `add_alg`)
                ocp.set_der(x1, (sz*z * (s1*x1) - (s2*x2) + su*u)/s1, scale=1/s1)
                ocp.set_der(x2, s1*x1/s2, scale=1/s2)

                # Lagrange objective term: signals in an integrand
                ocp.add_objective(ocp.integral(si*((s1*x1)**2 + (s2*x2)**2 + (su*u)**2))/si)
                # Mayer objective term: signals evaluated at t_f = t0_+T
                ocp.add_objective(ocp.at_tf((s1*x1)**2))

                # Path constraints
                #  (must be valid on the whole time domain running from `t0` to `tf`,
                #   grid options available such as `grid='inf'`)


                ocp.subject_to(s*s1*x1 >= -0.25*s, scale=s)
                ocp.subject_to(-1*s <= (su*u*s <= 1*s ), scale=s)

                # Boundary constraints
                ocp.subject_to(ocp.at_t0(s1*x1*s) == 0*s, scale=s)
                ocp.subject_to(ocp.at_t0(s2*x2*s) == 1*s, scale=s)

                #%%
                # Solving the problem
                # -------------------

                # Pick an NLP solver backend
                #  (CasADi `nlpsol` plugin):
                ocp.solver('sqpmethod',{"qpsol_options":{"print_problem":True,"dump_in":True,"dump":True},"convexify_strategy":"eigen-reflect"})


                # Pick a solution method
                #  e.g. SingleShooting, MultipleShooting, DirectCollocation
                #
                #  N -- number of control intervals
                #  M -- number of integration steps per control interval
                #  grid -- could specify e.g. UniformGrid() or GeometricGrid(4)
                method = Method(N=4)
                ocp.method(method)

                # Set initial guesses for states, controls and variables.
                #  Default: zero
                ocp.set_initial(u, linspace(0, 1, 4)/su) # Matrix

                # Solve
                sol = ocp.solve()

                f = cs.Function.load("qpsol.casadi")
                data=f.convert_in(f.generate_in("qpsol.000000.in.txt"))
                if run_type=="ref":
                    data_ref = data
                else:
                    for k in data.keys():

                        n = cs.norm_inf(data[k]-data_ref[k])
                        print(k, n)
                        assert n<=1e-7 # stems from integrator accuracy

        for e in glob.glob("qpsol.*"):
            os.remove(e)

if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#
#   Author: Jonathan Frey: jonathanpaulfrey(at)gmail.com

from casadi import n_nodes
import numpy as np


def structure_detection_print_summary(gnsf, acados_ocp):

    ## Description
    # this function prints the most important info after determining a GNSF
    # reformulation of the implicit model "initial_model" into "gnsf", which is
    # equivalent to the "reordered_model".
    model = acados_ocp.model
    # # GNSF
    # get dimensions
    nx = gnsf["nx"]
    nu = gnsf["nu"]
    nz = gnsf["nz"]

    nx1 = gnsf["nx1"]
    nx2 = gnsf["nx2"]

    nz1 = gnsf["nz1"]
    nz2 = gnsf["nz2"]

    # np = gnsf['np']
    n_out = gnsf["n_out"]
    ny = gnsf["ny"]
    nuhat = gnsf["nuhat"]

    #
    f_impl_expr = model.f_impl_expr
    n_nodes_initial = n_nodes(model.f_impl_expr)
    # x_old = model.x
    # f_impl_old = model.f_impl_expr

    x = gnsf["x"]
    z = gnsf["z"]

    phi_current = gnsf["phi_expr"]

    ## PRINT SUMMARY -- STRUCHTRE DETECTION
    print(" ")
    print(
        "*********************************************************************************************"
    )
    print(" ")
    print(
        "******************        SUCCESS: GNSF STRUCTURE DETECTION COMPLETE !!!      ***************"
    )
    print(" ")
    print(
        "*********************************************************************************************"
    )
    print(" ")
    print(
        f"========================= STRUCTURE DETECTION SUMMARY ===================================="
    )
    print(" ")
    print("-------- Nonlinear Static Feedback type system --------")
    print(" ")
    print(" successfully transcribed dynamic system model into GNSF structure ")
    print(" ")
    print(
        "reduced dimension of nonlinearity phi from        ",
        str(nx + nz),
        " to ",
        str(gnsf["n_out"]),
    )
    print(" ")
    print(
        "reduced input dimension of nonlinearity phi from  ",
        2 * nx + nz + nu,
        " to ",
        gnsf["ny"] + gnsf["nuhat"],
    )
    print(" ")
    print(f"reduced number of nodes in CasADi expression of nonlinearity phi from  {n_nodes_initial}  to  {n_nodes(phi_current)}\n")
    print("----------- Linear Output System (LOS) ---------------")
    if nx2 + nz2 > 0:
        print(" ")
        print(f"introduced Linear Output System of size           ", str(nx2 + nz2))
        print(" ")
        if nx2 > 0:
            print("consisting of the states:")
            print(" ")
            print(x[range(nx1, nx)])
            print(" ")
        if nz2 > 0:
            print("and algebraic variables:")
            print(" ")
            print(z[range(nz1, nz)])
            print(" ")
        if gnsf["purely_linear"] == 1:
            print(" ")
            print("Model is fully linear!")
            print(" ")
    if not all(gnsf["idx_perm_x"] == np.array(range(nx))):
        print(" ")
        print(
            "--------------------------------------------------------------------------------------------------"
        )
        print(
            "NOTE: permuted differential state vector x, such that x_gnsf = x(idx_perm_x) with idx_perm_x ="
        )
        print(" ")
        print(gnsf["idx_perm_x"])
    if nz != 0 and not all(gnsf["idx_perm_z"] == np.array(range(nz))):
        print(" ")
        print(
            "--------------------------------------------------------------------------------------------------"
        )
        print(
            "NOTE: permuted algebraic state vector z, such that z_gnsf = z(idx_perm_z) with idx_perm_z ="
        )
        print(" ")
        print(gnsf["idx_perm_z"])
    if not all(gnsf["idx_perm_f"] == np.array(range(nx + nz))):
        print(" ")
        print(
            "--------------------------------------------------------------------------------------------------"
        )
        print(
            "NOTE: permuted rhs expression vector f, such that f_gnsf = f(idx_perm_f) with idx_perm_f ="
        )
        print(" ")
        print(gnsf["idx_perm_f"])
    ## print GNSF dimensions
    print(
        "--------------------------------------------------------------------------------------------------------"
    )
    print(" ")
    print("The dimensions of the GNSF reformulated model read as:")
    print(" ")
    # T_dim = table(nx, nu, nz, np, nx1, nz1, n_out, ny, nuhat)
    # print( T_dim )
    print(f"nx    ", {nx})
    print(f"nu    ", {nu})
    print(f"nz    ", {nz})
    # print(f"np    ", {np})
    print(f"nx1   ", {nx1})
    print(f"nz1   ", {nz1})
    print(f"n_out ", {n_out})
    print(f"ny    ", {ny})
    print(f"nuhat ", {nuhat})



#! /usr/bin/python

import subprocess as sp
import jinja2 as jn
import sys
import json
import argparse
import re
from hashlib import sha1
from pathlib import Path
from collections import OrderedDict
import shutil
import multiprocessing



BLASFEO_PATH=str(Path(__file__).absolute().parents[1])
BLASFEO_TEST_PATH=str(Path(__file__).absolute().parents[0])

TEST_SCHEMA="test_schema.json"
TESTSET_JSON="testset_default.json"
BUILDS_DIR="build"
REPORTS_DIR="reports"
TESTCLASSES_DIR="classes"
TPL_PATH="Makefile.tpl"
LIB_BLASFEO_STATIC = "libblasfeo.a"
#MAKE_FLAGS={"--jobs":"8"}
MAKE_FLAGS={"--jobs": multiprocessing.cpu_count()}
SILENT=0



def parse_arguments():
	parser = argparse.ArgumentParser(description='BLASFEO tests scheduler')

	parser.add_argument(dest='testset_json', type=str, default=TESTSET_JSON, nargs='?',
		help='Run a batch of test from a specific testset, i.e. testset_all.json')
	parser.add_argument('--silent', default=False, action='store_true',
		help='Silent makefile output')
	parser.add_argument('--continue', dest='continue_test', default=False, action='store_true',
		help='Do not interrupt tests sweep on error')
	parser.add_argument('--rebuild', default=False, action='store_true',
		help='Rebuild libblasfeo to take into account recent code '+
		'changes or addition of new target to the testset batch')

	args = parser.parse_args()
	return args



def run_cmd(cmd, stdinput=""):

	if not SILENT: print(cmd,"\n")

	cmd_proc = sp.Popen(cmd,
		shell=True,
		stdin=sp.PIPE,
		stdout=sp.PIPE,
		stderr=sp.PIPE
		)

	outs, errs = cmd_proc.communicate(stdinput.encode("utf8"))

	if outs and not SILENT: print("Make infos:\n{}".format(outs.decode("utf8")))
	if errs: print("Make errors:\n{}".format(errs.decode("utf8")))

	return cmd_proc



def make_blasfeo(cmd="", env_flags={}, blasfeo_flags={}):

	# blasfeo compilation flags
	blasfeo_flags_str = " ".join([
		"{k}={v}".format(k=k, v=v)
		if v is not None else "{k}".format(k=k)
		for k, v in blasfeo_flags.items()
	])

	# enviroment flags
	env_flags_str = " ".join([
		"{k}={v}".format(k=k, v=v)
		for k, v in env_flags.items()
	])

	# compiler flags
	make_flags_str = " ".join([
		"{k}={v}".format(k=k, v=v)
		if v is not None else "{k}".format(k=k)
		for k, v in MAKE_FLAGS.items()
	])

	# clean build folder
	clean_cmd = "make deep_clean -C .. "
	cmd_proc = run_cmd(clean_cmd)

	# make blasfeo with selected flag set
	make_cmd = "{env_flags} make {blasfeo_flags} {make_flags} -C .. static_library"\
		.format(env_flags=env_flags_str,
				blasfeo_flags=blasfeo_flags_str,
				make_flags=make_flags_str, cmd=cmd)
	cmd_proc = run_cmd(make_cmd)

	return cmd_proc.returncode



class BlasfeoTestset:
	def __init__(self, cli_flags):
		global SILENT

		self.cli_flags=cli_flags
		self.continue_test = 0

		self.scheduled_routines = {}
		self.test_routines = {}

		with open(cli_flags.testset_json) as f:
			self.specs = json.load(f, object_pairs_hook=OrderedDict)

		with open(TPL_PATH) as f:
			self.makefile_template = jn.Template(f.read())

		if self.specs["options"].get("silent") or self.cli_flags.silent:
			SILENT = 1

		if self.specs["options"].get("continue") or  self.cli_flags.continue_test:
			self.continue_test = 1


		with open(TEST_SCHEMA) as f:
			self.schema = json.load(f, object_pairs_hook=OrderedDict)

		self._success_n = 0
		self._errors_n = 0

		self._total_n =\
			len(set(self.specs["routines"]))\
			* len(self.specs["TARGET"])\
			* len(self.specs["K_MAX_STACK"])\
			* len(self.specs["PACKING_ALG"])\
			* len(self.specs["precisions"])\
			* len(self.specs["apis"])

		# build standard testset skelethon
		self.build_testset()

	def parse_routine_options(self, routine_name, available_flags):

		# routine without any flag
		if not available_flags:
			return {'routine_basename': routine_name}

		pattern = '(?P<routine_basename>[a-z]*)_'

		for flag_name, flags_values in available_flags.items():
			flags_values = '|'.join(flags_values)
			pattern += '(?P<{flag_name}>[{flags_values}])'.format(flag_name=flag_name, flags_values=flags_values)

		parsed_flags = re.search(pattern, routine_name)

		if not parsed_flags:
			print("Error parsing flags of routine: {routine_name}".format(routine_name=routine_name))
			return {}

		return parsed_flags.groupdict()

	def build_testset(self):
		scheduled_routines = set(self.specs['routines'])

		# create testset with no global flags
		self.testset = OrderedDict(self.specs)
		self.testset["scheduled_routines"] = {}

		available_groups = self.schema['routines']

		# routine groups: blas1 blas2 ..
		for group_name, available_classes in available_groups.items():

			# routine classes: gemm, trsm, ...
			for class_name, routine_class in available_classes.items():
				available_routines = routine_class["routines"]
				routine_flags = routine_class["flags"]

				# routines: gemm, gemm_nn, gemm_nt, ...
				for routine  in available_routines:

					if routine not in scheduled_routines:
						continue

					scheduled_routines = scheduled_routines - {routine}

					# precision
					for precision in self.specs["precisions"]:

						# apis
						for api in self.specs["apis"]:

							test_macros = {}

							test_macros["ROUTINE_CLASS"] = class_name
							test_macros["PRECISION_{}".format(precision.upper())] = None

							routine_fullname = "{api}_{precision}{routine}".format(
								api=api, precision=precision[0], routine=routine)

							if api=="blas":
								test_macros["TEST_BLAS_API"] = None
								routine_dict = self.parse_routine_options(routine, routine_flags)
								if not routine_dict: continue
								test_macros.update(routine_dict)
								routine_testclass_src = "blasapi_"+routine_class["testclass_src"]
								routine_name = "{precision}{routine}".format(precision=precision[0], routine=class_name)
							else:
								routine_testclass_src = routine_class["testclass_src"]
								routine_name = "{precision}{routine}".format(precision=precision[0], routine=routine)

							test_macros["ROUTINE_CLASS_C"] = str(Path(TESTCLASSES_DIR, routine_testclass_src))
							test_macros["ROUTINE"] = routine_name
							test_macros["ROUTINE_FULLNAME"] = routine

							# add blas_api flag arguments values

							self.testset["scheduled_routines"][routine_fullname] = {
								"group": group_name,
								"class": class_name,
								"api": api,
								"precision": precision,
								"make_cmd": "update",
								"fullname": routine_fullname,
								"test_macros": test_macros
							}

		if scheduled_routines:
			print("Some routines were not found in the schema ({}) {}"
				  .format(TEST_SCHEMA, scheduled_routines))

	def run_all(self):
		# tune the testset and run

		for la in self.specs["LA"]:
			self.testset["blasfeo_flags"]["LA"]=la
			self.testset["test_macros"]["BLASFEO_LA"]=la

			if la=="REFERENCE":
				self.run_testset()
				break

			if la=="EXTERNAL_BLAS_WRAPPER":
				self.run_testset()
				break

			for mf in self.specs["MF"]:
				self.testset["blasfeo_flags"]["MF"]=mf
				self.testset["test_macros"]["BLASFEO_MF"]=mf

				for target in self.specs["TARGET"]:
					self.testset["blasfeo_flags"]["TARGET"]=target
					self.testset["test_macros"]["BLASFEO_TARGET"]=target

					for max_stack in self.specs["K_MAX_STACK"]:
						self.testset["blasfeo_flags"]["K_MAX_STACK"]=max_stack

						print("\n## Testing {la}:{target} kswitch={max_stack}".format(target=target, la=la, max_stack=max_stack))

						for alg in self.specs["PACKING_ALG"]:
							self.testset["blasfeo_flags"]["PACKING_ALG"]=alg
							self.testset["test_macros"]["PACKING_ALG"]=alg

							self.run_testset()

	def is_lib_updated(self):

		target = self.testset["blasfeo_flags"]["TARGET"]
		la = self.testset["blasfeo_flags"]["LA"]

		if self.lib_static_dst.is_file():
			return 1

		return 0

	def run_testset(self):
		# preparation step
		test_macros = self.testset["test_macros"]
		blasfeo_flags = self.testset["blasfeo_flags"]
		env_flags = self.testset["env_flags"]

		# always compile blas api
#		blasfeo_flags.update({"BLAS_API":1})
		blasfeo_flags.update({"BLASFEO_PATH":str(BLASFEO_PATH)})

		blasfeo_flags_str = "_".join(
			["{k}={v}".format(k=k, v=v) if v is not None else '' for k, v in blasfeo_flags.items()])

		# create unique folder name from set of flags
		m = sha1(blasfeo_flags_str.encode("utf8"))
		binary_dir = m.hexdigest()
		# binary path for current_conf build
		binary_path = Path(BLASFEO_TEST_PATH, BUILDS_DIR, binary_dir)
		# create directory
		binary_path.mkdir(parents=True, exist_ok=True)
		# update binary_dir flag
		blasfeo_flags.update({"ABS_BINARY_PATH":str(binary_path)})

		self.lib_static_src = Path(BLASFEO_PATH, "lib", LIB_BLASFEO_STATIC)

		self.lib_static_dst = Path(binary_path, LIB_BLASFEO_STATIC)

		lib_flags_json = str(Path(binary_path, "flags.json"))

		if self.cli_flags.silent or self.testset["options"].get("silent"):
			MAKE_FLAGS.update({"-s":None})

		if self.cli_flags.rebuild or self.testset["options"].get("rebuild") or not self.is_lib_updated():
			# compile the library
			make_blasfeo(blasfeo_flags=blasfeo_flags, env_flags=env_flags)

			# write used flags
			with open(lib_flags_json, "w") as f:
				json.dump(blasfeo_flags, f, indent=4)

			# copy library
			shutil.copyfile(str(self.lib_static_src), str(self.lib_static_dst))

			#  self.lib_static_dst.write_bytes(lib_static_src.read_bytes())

		for routine_name, args in self.testset['scheduled_routines'].items():
			# update local flags with global flags

			if args.get("test_macros"):
				args["test_macros"].update(test_macros)
			else:
				args["test_macros"] = test_macros

			if self.continue_test:
				args["test_macros"].update({"CONTINUE_ON_ERROR":1})

			if args.get("env_flags"):
				args["env_flags"].update(env_flags)
			else:
				args["env_flags"] = env_flags

			if args.get("blasfeo_flags"):
				args["blasfeo_flags"].update(blasfeo_flags)
			else:
				args["blasfeo_flags"] = blasfeo_flags

			error =  self.run_routine(routine_name, args)

			if error and not self.continue_test:
				break

	def render_routine(self, make_cmd="", env_flags={}, test_macros={}, blasfeo_flags={},
					   **kargs):

		# compress run_id
		la=blasfeo_flags['LA']
		if la != "HIGH_PERFORMANCE": target=la

		run_id = "{target}_{routine_fullname}_stack{kstack}_pack{alg}"\
			.format(
				target=blasfeo_flags['TARGET'],
				routine_fullname=kargs["fullname"],
				kstack=blasfeo_flags['K_MAX_STACK'],
				alg=blasfeo_flags['PACKING_ALG']
				)

		print("Testing {run_id}".format(run_id=run_id))

		makefile = self.makefile_template.render(test_macros=test_macros)

		# flag to override default libblasfeo flags
		blasfeo_flags_cmd = " ".join(["{k}={v}"\
			.format(k=k, v=v) if v is not None else k for k, v in blasfeo_flags.items()])
		make_flags = " ".join(["{k}={v}"\
			.format(k=k, v=v) if v is not None else k for k, v in MAKE_FLAGS.items()])

		make_cmd = "make {blasfeo_flags_cmd} -f - {make_cmd}"\
			.format(make_cmd=make_cmd, blasfeo_flags_cmd=blasfeo_flags_cmd)

		# render command to write on reports folder for later inspection
		report_cmd = "make {blasfeo_flags_cmd} {make_flags} {make_cmd}"\
			.format(make_cmd=make_cmd,
					make_flags=make_flags,
					blasfeo_flags_cmd=blasfeo_flags_cmd)

		# add entry in tested_routine
		self.test_routines[run_id] = {}
		self.test_routines[run_id]["make_cmd"] = make_cmd
		self.test_routines[run_id]["report_cmd"] = report_cmd
		self.test_routines[run_id]["makefile"] = makefile

		return run_id

	def run_routine(self, routine_fullname, kargs):

		run_id = self.render_routine(**kargs)

		make_cmd = self.test_routines[run_id]["make_cmd"]
		report_cmd = self.test_routines[run_id]["report_cmd"]
		makefile = self.test_routines[run_id]["makefile"]

		cmd_proc = run_cmd(make_cmd, stdinput=makefile)

		if cmd_proc.returncode:
			self._errors_n += 1
			self.test_routines[run_id]["success"]=0

			# on error write report for future inspection
			report_path = Path(REPORTS_DIR, run_id)
			report_path.mkdir(parents=True, exist_ok=True)
			with open(str(Path(report_path, "Makefile")), "w") as f:
				f.write(makefile)
			with open(str(Path(report_path, "make.sh")), "w") as f:
				f.write("#! /bin/bash\n")
				f.write(report_cmd+"\n")

			print("Error with {run_id}".format(run_id=run_id))

		else:
			self.test_routines[run_id]["success"]=1
			self._success_n += 1
			if not SILENT: print("Tested with {run_id}".format(run_id=run_id))

		# print partial
		if not SILENT: print("({done}:Succeded, {errors}:Errors) / ({total}:Total)\n"
			.format(done=self._success_n, errors=self._errors_n, total=self._total_n))

		return cmd_proc.returncode

	def print_summary(self):

		print("\n\n### Testset Summary:\n")
		summary_lines = [
			"Error with {run_id}".format(run_id=run_id)
			for (run_id, values) in self.test_routines.items() if not values["success"]
		]

		print("\n".join(summary_lines))
		print("({done}:Succeded, {errors}:Errors) / ({total}:Total)\n"
			.format(done=self._success_n, errors=self._errors_n, total=self._total_n))
		print("See blasfeo/tests/reports to inspect specific errors")


	def get_returncode(self):
		if self._errors_n > 0:
			return 1
		return 0



if __name__ == "__main__":

	cli_flags = parse_arguments()

	# generate test set
	# collection of routines/lib combinations to be run in the given excution of the tester.py
	testset = BlasfeoTestset(cli_flags)
	#  print(json.dumps(testset.testset, indent=4))
	testset.run_all()
	testset.print_summary()

	sys.exit(testset.get_returncode())
